/*
 * Zynaptic Stash - An asynchronous persistence framework for Java.
 * 
 * Copyright (c) 2009-2015, Zynaptic Limited. All rights reserved. 
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 only, as published by
 * the Free Software Foundation. Zynaptic Limited designates this particular
 * file as subject to the "Classpath" exception as provided in the LICENSE
 * file that accompanied this code.
 * 
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 * 
 * You should have received a copy of the GNU General Public License version 2
 * along with this work; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Please visit www.zynaptic.com or contact reaction@zynaptic.com if you need
 * additional information or have any questions.
 */

package com.zynaptic.stash.core;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;

/**
 * Encapsulates the serializable metadata associated with a given stash entry.
 * This can be written to the persistent store to allow the stash index to be
 * rebuilt on a warm start.
 * 
 * @param <T>
 *          This defines the serializable data type which may be stored for this
 *          stash entry.
 * 
 * @author Chris Holgate
 */
public class StashObjectMetadata<T extends Serializable> implements
    Serializable {
  private static final long serialVersionUID = -2352251054615619510L;

  // Define supported stash record formats.
  public static enum RecordFormat {
    PLACEHOLDER, RAW_OBJECT, TIME_SERIES
  };

  // Define stash entry metadata fields.
  private final RecordFormat recordFormat;
  private final String stashId;
  private final String stashName;
  private final byte[] encodedDefaultData;

  /**
   * Default constructor for building the stash entry metadata objects.
   * 
   * @param recordFormat
   *          This specifies the record format which is used to store the stash
   *          entry data.
   * @param stashId
   *          This is the unique identifier which is associated with the stash
   *          entry.
   * @param stashName
   *          This is the human readable name which is associated with the stash
   *          entry.
   * @param defaultData
   *          This is the default data value which is associated with the stash
   *          identifier.
   */
  public StashObjectMetadata(RecordFormat recordFormat, String stashId,
      String stashName, T defaultData) {
    this.recordFormat = recordFormat;
    this.stashId = stashId;
    this.stashName = stashName;

    // The default data value is serialised into a byte array so that the
    // metadata object can be processed without prior knowledge of its
    // associated data class loader.
    try {
      ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
      ObjectOutputStream objectStream = new ObjectOutputStream(byteStream);
      objectStream.writeObject(defaultData);
      objectStream.close();
      encodedDefaultData = byteStream.toByteArray();
    }

    // IO exceptions should never occur when writing to a local byte array, so
    // they are promoted to runtime exceptions here.
    catch (IOException error) {
      throw new RuntimeException(
          "Unexpected IO exception when writing to a local byte array.", error);
    }
  }

  /**
   * Gets the record format associated with the stash entry record.
   * 
   * @return Returns the record format which was used to store the stash entry
   *         data.
   */
  public final RecordFormat getRecordFormat() {
    return recordFormat;
  }

  /**
   * Gets unique identifier associated with the stash entry record.
   * 
   * @return Returns a string containing the unique identifier for the
   *         associated stash entry.
   */
  public final String getStashId() {
    return stashId;
  }

  /**
   * Gets human readable name associated with the stash entry record.
   * 
   * @return Returns a string containing the human readable name for the stash
   *         entry record.
   */
  public final String getStashName() {
    return stashName;
  }

  /**
   * Gets the default data value which is encoded in the stash entry metadata.
   * 
   * @return Returns a byte array containing the encoded default data value
   *         which was originally supplied when creating the stash entry
   *         metadata object.
   */
  public final byte[] getDefaultData() {
    return encodedDefaultData;
  }
}
