/*
 * Zynaptic Stash - An asynchronous persistence framework for Java.
 * 
 * Copyright (c) 2009-2015, Zynaptic Limited. All rights reserved. 
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 only, as published by
 * the Free Software Foundation. Zynaptic Limited designates this particular
 * file as subject to the "Classpath" exception as provided in the LICENSE
 * file that accompanied this code.
 * 
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 * 
 * You should have received a copy of the GNU General Public License version 2
 * along with this work; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Please visit www.zynaptic.com or contact reaction@zynaptic.com if you need
 * additional information or have any questions.
 */

package com.zynaptic.stash.osgi;

import java.util.logging.Level;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.osgi.util.tracker.ServiceTracker;
import org.osgi.util.tracker.ServiceTrackerCustomizer;

import com.zynaptic.reaction.Deferrable;
import com.zynaptic.reaction.Deferred;
import com.zynaptic.reaction.Logger;
import com.zynaptic.reaction.Reactor;
import com.zynaptic.stash.StashFactory;
import com.zynaptic.stash.core.StashFactoryCore;

/**
 * Implements default bundle activation for the Zynaptic Stash asynchronous
 * persistence service.
 * 
 * @author Chris Holgate
 */
public class StashDefaultActivator implements BundleActivator,
    ServiceTrackerCustomizer<Reactor, Reactor> {

  // Stash factory user API service registration.
  private ServiceRegistration<StashFactory> stashFactoryService;

  // Local context and bundle tracker references.
  private BundleContext context;
  private ServiceTracker<Reactor, Reactor> tracker;

  // Reaction client lifecycle handler.
  private LifecycleHandler lifecycleHandler;

  /**
   * Implementation of the <code> BundleActivator.start </code> method which is
   * used to start up the stash factory service on activation.
   * 
   * @param context
   *          The OSGi bundle context.
   */
  @Override
  public synchronized void start(BundleContext context) {
    this.context = context;
    tracker =
        new ServiceTracker<Reactor, Reactor>(context, Reactor.class.getName(),
            this);
    tracker.open();
  }

  /**
   * Implementation of the <code> ServiceTrackerCustomizer.addingService </code>
   * method which is used to indicate that the reactor service is available for
   * use by the client.
   * 
   * @param ref
   *          The service reference of the Reaction service.
   */
  @Override
  public synchronized Reactor addingService(ServiceReference<Reactor> ref) {
    Reactor reactor = null;

    // Multiple reactor services are not expected.
    if (lifecycleHandler == null) {
      reactor = context.getService(ref);
      lifecycleHandler = new LifecycleHandler();

      // Run the client lifecycle startup.
      lifecycleHandler.startup(reactor);

      // Register the reactor user interface as a provided service.
      stashFactoryService =
          context.registerService(StashFactory.class,
              lifecycleHandler.getStashFactory(), null);
    }

    // Return the reactor as the active service object.
    return reactor;
  }

  /**
   * Implementation of the <code> BundleActivator.stop </code> method which is
   * used to stop the stash factory service on deactivation.
   * 
   * @param context
   *          The OSGi bundle context.
   */
  @Override
  public synchronized void stop(BundleContext context) {

    // Perform orderly shutdown of the client before removing reactor
    // access.
    if (lifecycleHandler != null) {
      lifecycleHandler.teardown();
      lifecycleHandler = null;
      stashFactoryService.unregister();
      stashFactoryService = null;
    }
    tracker.close();
  }

  /**
   * Implementation of the
   * <code> ServiceTrackerCustomizer.removedService </code> method which is used
   * to indicate removal of the Reaction service. If called outside of an
   * ordered shutdown will abort the client due to loss of reactor
   * functionality.
   * 
   * @param ref
   *          The service reference of the Reaction service.
   * @param service
   *          The service object for the modified service.
   */
  @Override
  public synchronized void removedService(ServiceReference<Reactor> ref,
      Reactor service) {

    // Abort the client service if it is still running.
    if (lifecycleHandler != null) {
      lifecycleHandler.abort();
      lifecycleHandler = null;
      stashFactoryService.unregister();
      stashFactoryService = null;
    }
  }

  /**
   * Implementation of the
   * <code> ServiceTrackerCustomizer.modifiedService </code> method which is
   * used to indicate changes to the properties of the attached service. Since
   * the Reaction service does not support dynamic property changes, this
   * callback should never be invoked.
   * 
   * @param ref
   *          The service reference of the Reaction service.
   * @param service
   *          The service object for the modified service.
   */
  @Override
  public synchronized void modifiedService(ServiceReference<Reactor> ref,
      Reactor service) {
    service
        .getLogger("com.zynaptic.stash.osgi")
        .log(Level.SEVERE,
            "The Reaction service should not normally perform service modifications.");
  }

  /*
   * Reaction client lifecycle handler for OSGi Stash service.
   */
  private class LifecycleHandler {
    private final StashFactoryCore stashFactoryCore = new StashFactoryCore();
    private Logger logger = null;
    private volatile Thread waitingThread = null;

    /*
     * Gets a handle on the public stash factory interface.
     */
    private StashFactory getStashFactory() {
      return stashFactoryCore;
    }

    /*
     * Initialise the Reaction client on startup.
     */
    private void startup(Reactor reactor) {
      Thread localWaitingThread = waitingThread;
      if (localWaitingThread == null) {
        this.logger = reactor.getLogger("com.zynaptic.stash.osgi");
        waitingThread = Thread.currentThread();
        stashFactoryCore.setup(reactor).addDeferrable(
            new StartupCompleteHandler(), true);
        while (waitingThread != null) {
          try {
            Thread.sleep(60000);
          } catch (InterruptedException e) {
            // Interrupted exception indicates setup is complete.
          }
        }
      }
      System.out.println("Stash factory service brought up...");
    }

    /*
     * Tear down the Reaction client on orderly shutdown.
     */
    private void teardown() {
      Thread localWaitingThread = waitingThread;
      if (localWaitingThread == null) {
        waitingThread = Thread.currentThread();
        stashFactoryCore.teardown().addDeferrable(
            new TeardownCompleteHandler(), true);
        while (waitingThread != null) {
          try {
            Thread.sleep(60000);
          } catch (InterruptedException e) {
            // Interrupted exception indicates teardown is complete.
          }
        }
      }
      System.out.println("Stash factory service torn down...");
    }

    /*
     * Abort the Reaction client on loss of reactor service.
     */
    private void abort() {
      System.err
          .println("Stash service aborted on unexpected loss of reactor.");
      // TODO: Clean aborts are not currently supported.
    }

    /*
     * Callback on completion of stash factory startup.
     */
    private final class StartupCompleteHandler implements
        Deferrable<Boolean, Boolean> {
      @Override
      public Boolean onCallback(Deferred<Boolean> deferred, Boolean status) {
        Thread localWaitingThread = waitingThread;
        if (localWaitingThread != null) {
          logger.log(Level.INFO, "Started stash factory service.");
          waitingThread = null;
          localWaitingThread.interrupt();
        }
        return null;
      }

      @Override
      public Boolean onErrback(Deferred<Boolean> deferred, Exception error) {
        Thread localWaitingThread = waitingThread;
        if (localWaitingThread != null) {
          logger.log(Level.SEVERE, "Failed to start stash factory service.",
              error);
          waitingThread = null;
          localWaitingThread.interrupt();
        }
        return null;
      }
    }

    /*
     * Callback on completion of stash factory startup.
     */
    private final class TeardownCompleteHandler implements
        Deferrable<Boolean, Boolean> {
      @Override
      public Boolean onCallback(Deferred<Boolean> deferred, Boolean status) {
        Thread localWaitingThread = waitingThread;
        if (localWaitingThread != null) {
          logger.log(Level.INFO, "Halted stash factory service.");
          waitingThread = null;
          localWaitingThread.interrupt();
        }
        return null;
      }

      @Override
      public Boolean onErrback(Deferred<Boolean> deferred, Exception error) {
        Thread localWaitingThread = waitingThread;
        if (localWaitingThread != null) {
          logger.log(Level.SEVERE, "Failed to halt stash factory service.",
              error);
          waitingThread = null;
          localWaitingThread.interrupt();
        }
        return null;
      }
    }
  }
}
